
-- The RTL description of a $K$ = 9 adaptive Viterbi decoder is given below. 
-- $N_{max}$ = 9
-- $TL$ = 45
-- $T$ = 18

-- Clock : CPE_Pclk
-- Reset : CPE_Reset

library IEEE;
use IEEE.std_logic_1164.all;
use ieee.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

-------------------------- Architecture Declaration --------------------------

architecture behavior of ava is

constant k : integer := 8;                                            -- K-1, where K is 9              
constant nmax : integer := 9;                                         -- Value of Nmax (Max number of paths allowed)
constant tl : integer := 45;                                          -- Truncation length 

type bm_array is array (3 downto 0) of std_logic_vector(3 downto 0);  -- defines output of branch metric unit
signal b : bm_array;                                                  -- branch metric outputs

signal initial, final : std_logic;                                    -- indicate initial and final input bit
                                                                      -- Russ: check on this.

type state_type is (s0, s1, s2, s3, s4, s5, s6, s7, s8, s9, s10,      -- decoder states
s11, s12, s13, dummy, stest, sriram, tintin, asterix, monkey,         -- Russ: develop better states
dumbo, s_add, Read1, Read2, jim, 
don, bob, tom, dummy_s3, 
dummy_tintin, dummy_s8, dummy_dumbo, dummy_s12, minimum);


signal present_state, next_state  : state_type;               -- indicates states

signal valid : std_logic_vector(nmax-1 downto 0);             -- indicates which paths in surv. mem are valid.
signal update1 : std_logic_vector(nmax-1 downto 0);
signal shift : std_logic_vector(k-1 downto 0);
signal dm : std_logic_vector(5 downto 0);
type op_array is array (nmax-1 downto 0) of std_logic_vector(1 downto 0);
signal output0, output1 : op_array; -- encoder output
type ps_array is array (nmax-1 downto 0) of std_logic_vector(k-1 downto 0);
signal presentstate : ps_array;                
signal ll : integer range 0 to 30;    
signal stud, honda : std_logic_vector(nmax-1 downto 0);
signal level : std_logic_vector(tl-1 downto 0); -- pointer to update survivor memory
type ns_array is array (nmax-1 downto 0) of std_logic_vector(k-1 downto 0);
signal ns0, ns1 : ns_array;
signal T : std_logic_vector(5 downto 0);  -- threshold
type next_array is array (nmax-1 downto 0) of std_logic_vector(k-1 downto 0);
signal NEXTSTATE0, NEXTSTATE1 : next_array;
type cs_array is array (nmax-1 downto 0) of std_logic_vector(5 downto 0);
signal d0, d1, add0, add1 : cs_array;
type csmain_array is array (nmax-1 downto 0) of std_logic_vector(5 downto 0);
signal PM, p_metric : csmain_array;
type index_ns is array (nmax-1 downto 0) of std_logic;
signal PathIdentify_0, PathIdentify_1, PathIdentify : index_ns;
--type inputs is array (14 downto 0) of std_logic_vector(2 downto 0);
--signal ip1_array, ip2_array : inputs;
type mem_array is array (nmax-1 downto 0) of std_logic_vector(tl-1 downto 0);
signal memory : mem_array;
signal calc_b, inc_T, dec_T, reset_T, inc_level, reset_level : std_logic; 
signal inc_ll, reset_ll, inc_add, reset_honda, inc_honda: std_logic;
signal inc_stud, reset_stud, inc_count, reset_count : std_logic;
signal ip1, ip2 : std_logic_vector(2 downto 0);
signal out0, out1 : std_logic_vector(nmax-1 downto 0);
signal temp : std_logic_vector(5 downto 0);
type min_array is array (nmax-1 downto 0) of std_logic_vector(5 downto 0);
signal min1: min_array;
type min2_array is array (3 downto 0) of std_logic_vector(5 downto 0);
signal min2 : min2_array;
type row_array is array (3 downto 0) of std_logic_vector(nmax-1 downto 0);
signal row1 : row_array;
type min3_array is array (1 downto 0) of std_logic_vector(5 downto 0);
signal min3 : min3_array;
type row2_array is array (1 downto 0) of std_logic_vector(nmax-1 downto 0);
signal row2 : row2_array;
signal count : integer range 0 to 50;
signal inc_bun, inc_bb, reset_bun, reset_bb, norm_metric : std_logic; 
signal set_out, set_ns, set_output : std_logic;
signal read_ip, set_cumsum_s7,set_cumsum_s8,set_cumsum_don : std_logic;
signal reset_valid, reset_xy, set_x, set_y : std_logic;
signal x, y, bun, bb, h : std_logic_vector(nmax-1 downto 0);
signal set_min1, set_min2, set_min3, set_dm, reset_dm : std_logic;
signal set_h, reset_h, set_fifo_out, setpath_don : std_logic;
signal resetcsumpath, setcsum_sriram, setcsum_tintin : std_logic; 
signal setcsum_monkey, setcsum_asterix : std_logic;
signal reset_row, reset_nextstates : std_logic;

begin  -- decoder
         -- purpose: Branch metric generator
         -- type   : sequential
         -- inputs : clock, reset
         -- outputs: branch metrics
         bmg : process (CPE_Pclk,CPE_Reset)

         begin  -- process bmg
           if (CPE_Reset ='1') then
              for p in 3 downto 0 loop
                   b(p) <= (others => '0');
              end loop;  -- i
           elsif (CPE_Pclk'event and CPE_Pclk = '1') then
             if (calc_b = '1') then
               b(0) <= ('0' & ip1) + ('0' & ip2);
               b(1) <= ('0' & ip1) + ("0111"-('0' & ip2)); 
               b(2) <= ('0' & ip2) + ("0111"-('0' & ip1));  
               b(3) <= ("0111"- ('0' & ip1)) + ("0111"- ('0' & ip2));
             end if;
           end if; 
         end process bmg;

         -- purpose: RESET
         -- type   : sequential
         -- inputs : clock, reset
         -- outputs: 
         state_clocked : process (CPE_Pclk, CPE_Reset)
         begin  -- process
           if CPE_Reset = '1' then
             present_state <= s0;
           elsif CPE_Pclk'event and CPE_Pclk = '1' then
             present_state <= next_state;
           end if;
         end process state_clocked;
         
         -- purpose: Threshold Update
         -- type   : sequential
         -- inputs : CPE_Pclk, CPE_Reset
         -- outputs: T
         T_update: process (CPE_Pclk, CPE_Reset)
         begin  -- process T_update
           if CPE_Reset = '1' then      -- asynchronous reset (active low)
             T <= "010010"; -- T = 18
           elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
             if dec_T = '1' then
               T <= T - 2;
             elsif inc_T = '1' then
               T <= T + 2;
             elsif reset_T = '1' then
               T <= "010100";
             end if;
           end if;
         end process T_update;
         
         -- purpose: Level Update
         -- type   : sequential
         -- inputs : CPE_Pclk, CPE_Reset
         -- outputs: level
         level_update: process (CPE_Pclk, CPE_Reset)
         begin  -- process level
           if CPE_Reset = '1' then      -- asynchronous reset (active low)
              level(tl-1 downto 1) <= (others => '0');
              level(0) <= '1';
            -- level <= 0;
           elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
             if reset_level = '1' then
                level(tl-1 downto 1) <= (others => '0');
                level(0) <= '1';
--               level <= 0;
             elsif inc_level = '1' then
               level <= level(tl-2 downto 0) & level(tl-1);
               --level <= level + 1;
             end if;
           end if;
         end process level_update;

         -- purpose: ll update
         -- type   : sequential
         -- inputs : CPE_Pclk, CPE_Reset
         -- outputs: ll
         ll_update: process (CPE_Pclk, CPE_Reset)
         begin  -- process ll_update
           if CPE_Reset = '1' then      -- asynchronous reset (active low)
             ll <= 0;
           elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
             if inc_ll = '1' then
               ll <= ll + 1;
             elsif reset_ll = '1' then
               ll <= 0;
             end if;
           end if;
         end process ll_update;

         -- purpose: path metrics
         -- type   : sequential
         -- inputs : CPE_Pclk, CPE_Reset
         -- outputs: branch metrics
         metrics: process (CPE_Pclk, CPE_Reset)
         begin  -- process hamming
           if CPE_Reset = '1' then      -- asynchronous reset (active low)
             add0 <= (others => "000000");
             add1 <= (others => "000000");
           elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
             if inc_add = '1' then
               for i in 0 to nmax-1 loop
               if valid(i) = '1' then
                 case output0(i) is
                   when "00" =>
                     add0(i) <= ("00" & b(0)) + PM(i);
                   when "01" =>
                     add0(i) <= ("00" & b(1)) + PM(i);
                   when "10" =>
                     add0(i) <= ("00" & b(2)) + PM(i);
                   when "11" =>
                     add0(i) <= ("00" & b(3)) + PM(i);
                   when others => null;
                 end case;
                 case output1(i) is
                   when "00" =>
                     add1(i) <= ("00" & b(0)) + PM(i);
                   when "01" =>
                     add1(i) <= ("00" & b(1)) + PM(i);
                   when "10" =>
                     add1(i) <= ("00" & b(2)) + PM(i);
                   when "11" =>
                     add1(i) <= ("00" & b(3)) + PM(i);
                   when others => null;
                 end case;
               end if;
               end loop;
             end if;
           end if;
         end process metrics;

         -- purpose: control honda
         -- type   : sequential
         -- inputs : CPE_Pclk, CPE_Reset
         -- outputs: honda
         running_out_of_names: process (CPE_Pclk, CPE_Reset)
         begin  -- process running_out_of_names
           if CPE_Reset = '1' then      -- asynchronous reset (active low)
             honda(nmax-1 downto 1) <= (others => '0');
             honda(0) <= '1';
           elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
             if reset_honda = '1' then
               honda(nmax-1 downto 1) <= (others => '0');
               honda(0) <= '1';
             elsif inc_honda = '1' then
               honda <= honda(nmax-2 downto 0) & '0';
             end if;
           end if;
         end process running_out_of_names;

         
         -- purpose:
         -- 
         count_update: process (CPE_Pclk, CPE_Reset)
         begin  -- process count update
           if CPE_Reset = '1' then      -- asynchronous reset (active low)
             count <= 0;
           elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
             if reset_count = '1' then
               count <= 0;
             elsif inc_count = '1' then
               count <= count + 1;
             end if;
           end if;
         end process count_update;

         -- purpose: stud increment
         -- type   : sequential
         -- inputs : CPE_Pclk, CPE_Reset
         -- outputs: stud
         stud_update: process (CPE_Pclk, CPE_Reset)
         begin  -- process stud_update
           if CPE_Reset = '1' then      -- asynchronous reset (active low)
             stud(nmax-1 downto 1) <= (others => '0');
             stud(0) <= '1';                         
           elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
             if reset_stud = '1' then
               stud(nmax-1 downto 1) <= (others => '0');
               stud(0) <= '1';
             elsif inc_stud = '1' then
               stud <= stud(nmax-2 downto 0) & '0';
             end if;
           end if;
         end process stud_update;

        
         -- purpose: update bun
         -- type   : sequential
         -- inputs : CPE_Pclk, CPE_Reset
         -- outputs: bun
         bunupdate: process (CPE_Pclk, CPE_Reset)
         begin  -- process bunupdate
           if CPE_Reset = '1' then      -- asynchronous reset (active low)
             bun(nmax-1 downto 1) <= (others => '0');
             bun(0) <= '1';
           elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
             if reset_bun = '1' then
               --bun <= 0;
               bun(nmax-1 downto 1) <= (others => '0');
               bun(0) <= '1';
             elsif inc_bun = '1' then
               --bun <= bun + 1;
               bun <= bun(nmax-2 downto 0) & '0';
             end if;
           end if;
         end process bunupdate;

         -- purpose: update bb
         -- type   : sequential
         -- inputs : CPE_Pclk, CPE_Reset
         -- outputs: bb
         bbupdate: process (CPE_Pclk, CPE_Reset)
         begin  -- process bunupdate
           if CPE_Reset = '1' then      -- asynchronous reset (active low)
             bb(nmax-1 downto 1) <= (others => '0');
             bb(0) <= '1';
           elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
             if reset_bb = '1' then
               bb(nmax-1 downto 1) <= (others => '0');
               bb(0) <= '1';
             elsif inc_bb = '1' then
               bb <= bb(nmax-2 downto 0) & '0';
             end if;
           end if;
         end process bbupdate;

        
         -- purpose: Convolutional Encoder
         -- type   : sequential
         -- inputs : CPE_Pclk, CPE_Reset
         -- outputs: out0, out1 are the two outputs described as O0 and O1
         out_update: process (CPE_Pclk, CPE_Reset)
         begin  -- process bunupdate
           if CPE_Reset = '1' then      -- asynchronous reset (active low)
             out0 <= (others => '0');
             out1 <= (others => '0');   
           elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
             if set_out = '1' then
                for g in 0 to nmax-1 loop
                   if valid(g) = '1' then
                      out0(g) <= ('0' xor presentstate(g)(0)) xor (presentstate(g)(1) 
				  xor presentstate(g)(2) xor 
                                  presentstate(g)(3) xor presentstate(g)(4) xor 
				  presentstate(g)(5) xor presentstate(g)(6));
                      out1(g) <= ('1' xor presentstate(g)(0)) xor (presentstate(g)(1) 
				  xor presentstate(g)(2) xor presentstate(g)(3) 
				  xor presentstate(g)(4) xor presentstate(g)(5) 
				  xor presentstate(g)(6));
                   end if;
                end loop;
             end if;
           end if;
         end process out_update;

         -- purpose: output update
         -- type   : sequential
         -- inputs : CPE_Pclk, CPE_Reset
         -- outputs: output0, output1
         output_update: process (CPE_Pclk, CPE_Reset)
         begin  -- process bunupdate
           if CPE_Reset = '1' then      -- asynchronous reset (active low)
             output0 <= (others => "00");
             output1 <= (others => "00");       
           elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
             if set_output = '1' then
                for g in 0 to nmax-1 loop
                   if valid(g) = '1' then
                     output0(g) <=  '0' & out0(g);
                     output1(g) <=  '1' & out1(g);      
                   end if;
                end loop;
             end if;
           end if;
         end process output_update;


        -- purpose: nsupdate NEXT STATE COMPUTATION
         -- type   : sequential
         -- inputs : CPE_Pclk, CPE_Reset
         -- outputs: ns0, ns1
         ns_update: process (CPE_Pclk, CPE_Reset)
         begin  -- process bunupdate
           if CPE_Reset = '1' then      -- asynchronous reset (active low)
             ns0 <= (others => "00000000");
             ns1 <= (others => "00000000");  
           elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
             if set_ns = '1' then
                for g in 0 to nmax-1 loop
                   if valid(g) = '1' then
                      ns0(g) <= '0' & presentstate(g)(k-1 downto 1); -- shift register  
                      ns1(g) <= '1' & presentstate(g)(k-1 downto 1); -- shown in Figure 5.2    
                   end if;
                end loop;
             end if;
           end if;
         end process ns_update;
        

        -- purpose: get inputs
        -- type   : sequential
        -- inputs : CPE_Pclk, CPE_Reset
        -- outputs: ip1, ip2, initial, final
        dec_inputs: process (CPE_Pclk, CPE_Reset)
        begin  -- process inputs
          if CPE_Reset = '1' then       -- asynchronous reset (active low)
              ip1 <= (others => '0');
              ip2 <= (others => '0');
              initial <= '0';
              final <= '0';
          elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
            if (read_ip = '1') then
              ip1 <= CPE_HostToPeFifoData_In(2 downto 0); -- first 3-bit quantized input symbol
              ip2 <= CPE_HostToPeFifoData_In(5 downto 3); -- second 3-bit quantized input symbol
              initial <= CPE_HostToPeFifoData_In(6); -- indicates start of transmission
              final <= CPE_HostToPeFifoData_In(7); -- indicates end of transmission
            end if;
          end if;
        end process dec_inputs;


         -- purpose: PM update 
         -- type   : sequential
         -- inputs : CPE_Pclk, CPE_Reset
         -- outputs: PM
         update_sum: process (CPE_Pclk, CPE_Reset)
         begin  -- process update_sum
           if CPE_Reset = '1' then      -- asynchronous reset (active low)
             PM <= (others => "000000");
           elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
             if set_cumsum_s7 = '1' then
               for c in 0 to nmax-1 loop
                 if PathIdentify_0(c) = '1' then
                   if valid(c) = '0' then
                     PM(c) <= d0(c);
                   end if;
                 end if;
               end loop;  -- c
             elsif set_cumsum_s8 = '1' then
               for a in 0 to nmax-1 loop
                 if PathIdentify_1(a) = '1' then
                   if valid(a) = '0' then
                     PM(a) <= d1(a);
                   end if;
                 end if;
               end loop;  -- a
             elsif set_cumsum_don = '1' then
               for buddy in 0 to nmax-1 loop
                 if y(buddy) = '1' then
                   for kk in 0 to nmax-1 loop
                     if x(kk) = '1' then
                       PM(buddy) <= d1(kk);
                     end if;
                   end loop;
                 end if;
               end loop;
             elsif norm_metric = '1' then
               for dd in 0 to nmax-1 loop
                 if valid(dd) = '1' then
                   PM(dd) <= PM(dd) - dm; -- normalize path metrics 
                 end if;
               end loop;  -- dd    
             end if;
           end if;
           end process update_sum;


         -- purpose: valid update VALID BIT OF SURVIVOR MEMORY ROW
         -- type   : sequential
         -- inputs : CPE_Pclk, CPE_Reset
         -- outputs: valid
         update_valid: process (CPE_Pclk, CPE_Reset)
         begin  -- process update_sum
           if CPE_Reset = '1' then      -- asynchronous reset (active low)
             valid <=  "000000001";
           elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
             if set_cumsum_s7 = '1' then
               for c in 0 to nmax-1 loop
                 if PathIdentify_0(c) = '1' then
                   if valid(c) = '0' then
                     valid(c) <= '1'; -- indicates existence of a path in row c
                   end if;
                 end if;
               end loop;  -- c
             elsif set_cumsum_s8 = '1' then
               for a in 0 to nmax-1 loop
                 if PathIdentify_1(a) = '1' then
                   if valid(a) = '0' then
                     valid(a) <= '1';
                   end if;
                 end if;
               end loop;  -- a
             elsif set_cumsum_don = '1' then
               for buddy in 0 to nmax-1 loop
                 if y(buddy) = '1' then
                   for kk in 0 to nmax-1 loop
                     if x(kk) = '1' then
                      valid(buddy) <= '1';
                     end if;
                   end loop;
                 end if;
               end loop;
             elsif  reset_valid = '1' then
               valid <= (others => '0');    
             end if;
           end if;
           end process update_valid;

          -- purpose: presentstate update Store NEXTSTATE TO Present state
         -- type   : sequential
         -- inputs : CPE_Pclk, CPE_Reset
         -- outputs: presentstate
         update_presentstate: process (CPE_Pclk, CPE_Reset)
         begin  -- process update_sum
           if CPE_Reset = '1' then      -- asynchronous reset (active low)
             presentstate <= (others => "00000000");
           elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
             if set_cumsum_s7 = '1' then
               for c in 0 to nmax-1 loop
                 if PathIdentify_0(c) = '1' then
                   if valid(c) = '0' then
                     presentstate(c) <= NEXTSTATE0(c);
                   end if;
                 end if;
               end loop;  -- c
             elsif set_cumsum_s8 = '1' then
               for a in 0 to nmax-1 loop
                 if PathIdentify_1(a) = '1' then
                   if valid(a) = '0' then
                     presentstate(a) <= NEXTSTATE1(a);
                   end if;
                 end if;
               end loop;  -- a
             elsif set_cumsum_don = '1' then
               for buddy in 0 to nmax-1 loop
                 if y(buddy) = '1' then
                   for kk in 0 to nmax-1 loop
                     if x(kk) = '1' then
                       presentstate(buddy) <= NEXTSTATE1(kk);
                     end if;
                   end loop;
                 end if;
               end loop;
             end if;
           end if;
           end process update_presentstate;

         -- purpose: survivor memory update
         -- type   : sequential
         -- inputs : CPE_Pclk, CPE_Reset
         -- outputs: memory
         update_memory: process (CPE_Pclk, CPE_Reset)
         begin  -- process update_memory
           if CPE_Reset = '1' then      -- asynchronous reset (active low)
             memory <= (others => "000000000000000000000000000000000000000000000");
           elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
             if set_cumsum_s7 = '1' then
               for c in 0 to nmax-1 loop
                 if PathIdentify_0(c) = '1' then
                   if valid(c) = '0' then
                     for i in 0 to tl-1 loop
                       if level(i) = '1' then
                         memory(c)(i) <= NEXTSTATE0(c)(k-1);  -- Corresponds to Figure 5.11
	                 -- Decision bit = left-most bit of NEXTSTATE
                       end if;
                     end loop;  -- i
                   end if;
                 end if;
               end loop;  -- c
             elsif set_cumsum_s8 = '1' then
               for a in 0 to nmax-1 loop
                 if PathIdentify_1(a) = '1' then
                   if valid(a) = '0' then
                     for i in 0 to tl-1 loop
                       if level(i) = '1' then
                         memory(a)(i) <= NEXTSTATE1(a)(k-1);  --doubt     
                       end if;
                     end loop;  -- i
                   end if;
                 end if;
               end loop;  -- a
             elsif set_cumsum_don = '1' then
               for buddy in 0 to nmax-1 loop
                 if y(buddy) = '1' then
                   for kk in 0 to nmax-1 loop
                     if x(kk) = '1' then
                       for i in 0 to tl-1 loop
                         if level(i) = '1' then
                           memory(buddy)(i) <= NEXTSTATE1(kk)(k-1);
                         else
                           memory(buddy)(i) <= memory(kk)(i);
                         end if;
                       end loop;  -- i
                     end if;
                   end loop;
                 end if;
               end loop;
             end if;
           end if;
           end process update_memory;

           -- purpose: y -- Search for a row with valid = '0'
           -- type   : sequential
           -- inputs : CPE_Pclk, CPE_Reset
           -- outputs: y
           updatey: process (CPE_Pclk, CPE_Reset)
           begin  -- process updatey
             if CPE_Reset = '1' then    -- asynchronous reset (active low)
               y <= (others => '0');
             elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
               if reset_xy = '1' then
                 y <= (others => '0');
               elsif set_y = '1' then
                 for dolly in 0 to nmax-1 loop
                   if bb(dolly) = '1' then
                     if valid(dolly) = '0' then
                       y(dolly) <= '1';
                     end if;
                   end if;
                 end loop;
               end if;
             end if;
             end process updatey;

           -- purpose: x
           -- type   : sequential
           -- inputs : CPE_Pclk, CPE_Reset
           -- outputs: x
           updatex: process (CPE_Pclk, CPE_Reset)
           begin  -- process updatey
             if CPE_Reset = '1' then    -- asynchronous reset (active low)
               x <= (others => '0');
             elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
               if reset_xy = '1' then
                 x <= (others => '0');
               elsif set_x = '1' then
                 for pop in 0 to nmax-1 loop
                   if bun(pop) = '1' then
                     if update1(pop) = '1' then
                       x(pop) <= '1';
                     end if;
                   end if;
                 end loop;
               end if;
             end if;
             end process updatex;
           
             -- purpose: update update1
             -- type   : sequential
             -- inputs : CPE_Pclk, CPE_Reset
             -- outputs: update1
             update_update1: process (CPE_Pclk, CPE_Reset)
             begin  -- process update_update1
               if CPE_Reset = '1' then  -- asynchronous reset (active low)
                 update1 <= (others => '0');
               elsif CPE_Pclk'event and CPE_Pclk = '1' then  -- rising clock edge
                 if set_cumsum_s8 = '1' then
                   for a in 0 to nmax-1 loop
                     if PathIdentify_1(a) = '1' then
                       if valid(a) = '1' then
                         update1(a) <= '1';
                       end if;
                     end if;
                   end loop;  -- a
                 elsif set_cumsum_don = '1' then
                   for buddy in 0 to nmax-1 loop
                     if y(buddy) = '1' then
                       for kk in 0 to nmax-1 loop
                         if x(kk) = '1' then
                           update1(kk) <= '0';
                         end if;
                       end loop;
                     end if;
                   end loop;
                 end if;
               end if;
             end process update_update1;


             -- purpose: min1
             -- type   : sequential
             -- inputs : CPE_Pclk, CPE_Reset
             -- outputs: min1
             min1_update: process (CPE_Pclk, CPE_Reset)
             begin  -- process min1_update
               if CPE_Reset = '1' then  -- asynchronous reset (active low)
                 min1 <= (others => "000000");
               elsif CPE_Pclk'event and CPE_Pclk = '1' then  
                 if set_min1 = '1' then
                   for j in 0 to nmax-1 loop
                     if d0(j) < d1(j) then
                       min1(j) <= d0(j);
                     else
                       min1(j) <= d1(j);
                     end if;
                   end loop;
                 end if;
               end if;                  -- rising clock edge
             end process min1_update;


             -- purpose: min2
             -- type   : sequential
             -- inputs : CPE_Pclk, CPE_Reset
             -- outputs: min2
             min2_update: process (CPE_Pclk, CPE_Reset)
             begin  -- process min2_update
               if CPE_Reset = '1' then  -- asynchronous reset (active low)
                 min2 <= (others => "000000");
                 row1 <= (others => "000000000");
               elsif CPE_Pclk'event and CPE_Pclk = '1' then  
                 if set_min2 = '1' then
                   for j in 0 to 3 loop
                     if min1(j) < min1(nmax-1-j) then
                       min2(j) <= min1(j);
                       row1(j)(j) <= '1';
                     else
                       min2(j) <= min1(nmax-1-j);
                       row1(j)(nmax-1-j) <= '1';
                     end if;
                   end loop;  -- j
                 elsif reset_row = '1' then
                   row1 <= (others => "000000000");
                 end if;
               end if;
             end process min2_update;


             -- purpose: min3 update
             -- type   : sequential
             -- inputs : CPE_Pclk, CPE_Reset
             -- outputs: min3
             min3_update: process (CPE_Pclk, CPE_Reset)
             begin  -- process min3_update
               if CPE_Reset = '1' then  -- asynchronous reset (active low)
                 min3 <= (others => "000000");
                 row2 <= (others => "000000000");
               elsif CPE_Pclk'event and CPE_Pclk = '1' then  
                 if set_min3 = '1' then
                   for j in 0 to 1 loop
                     if min2(j) < min2(3-j) then
                       min3(j) <= min2(j);
                       row2(j) <= row1(j);
                     else
                       min3(j) <= min2(3-j);
                       row2(j) <= row1(3-j);
                     end if;
                   end loop;  -- j
                 elsif reset_row = '1' then
                   row2 <= (others => "000000000");
                 end if;
               end if;
             end process min3_update;
             
             -- purpose: dm
             -- type   : sequential
             -- inputs : CPE_Pclk, CPE_Reset
             -- outputs: dm
             dm_update: process (CPE_Pclk, CPE_Reset)
             begin  -- process min2_update
               if CPE_Reset = '1' then  -- asynchronous reset (active low)
                 dm <= "000000";
                 h <= (others => '0');
               elsif CPE_Pclk'event and CPE_Pclk = '1' then  
                 if set_dm = '1' then
                   if min3(0) < min3(1) then
                     dm <= min3(0);
                     h <= row2(0);
                   else
                     dm <= min3(1);
                     h <= row2(1);
                   end if;
                 elsif reset_dm = '1' then
                   dm <= "000000";
                   h <= (others => '0');
                 end if;
               end if;
             end process dm_update;

             -- purpose: fifo output
             -- type   : sequential
             -- inputs : CPE_Pclk, CPE_Reset
             -- outputs: fifo
             fifowrite: process (CPE_Pclk, CPE_Reset)
             begin  -- process fifowrite
               if CPE_Reset = '1' then  -- asynchronous reset (active low)
                 CPE_PeToHostFifoData_Out(0) <= '0';
               elsif CPE_Pclk'event and CPE_Pclk = '1' then  
                 if set_fifo_out = '1' then
                   if level(tl-1) = '0' then
                     for jj in 0 to tl-2 loop
                       if level(jj) = '1' then
                         for mmm in 0 to nmax-1 loop
                           if h(mmm) = '1' then
                             CPE_PeToHostFifoData_Out(0) <= memory(mmm)(jj+1);
                           end if;
                         end loop;  -- mmm
                         --CPE_PeToHostFifoData_Out(0) <= memory(h)(jj+1);
                       end if;
                     end loop;
                   else
                     for mmm in 0 to nmax-1 loop
                       if h(mmm) = '1' then
                         CPE_PeToHostFifoData_Out(0) <= memory(mmm)(0);
                       end if;
                     end loop;  -- mmm
                   end if;
                 end if;
               end if;
             end process fifowrite;
               

             csum0_update: process (CPE_Pclk, CPE_Reset)
             begin  -- process d0_update
               if CPE_Reset = '1' then  -- asynchronous reset (active low)
                 d0 <= (others => "111111");
               elsif CPE_Pclk'event and CPE_Pclk = '1' then  
                  if resetcsumpath = '1' then
                    d0 <= (others => "111111");
                  elsif setcsum_sriram = '1' then
                    for i in 0 to nmax-1 loop
                      if stud(i) = '1' then
                        if valid(i) = '1' then
                          if add0(i) < T then
                            d0(i) <= add0(i);
                          else
                            d0(i) <= (others => '1');
                          end if;
                        end if;
                      end if;
                    end loop;
                  elsif setcsum_tintin = '1' then
                    for i in 0 to nmax-1 loop
                      if honda(i) = '1' then
                        if valid(i) = '1' then
                          if add0(i) < T then
                            d0(i) <= add0(i);
                          else
                            d0(i) <= "111111";
                          end if;
                        end if;
                      end if;
                    end loop;
                  end if;
               end if;
             end process csum0_update;


             csum1_update: process (CPE_Pclk, CPE_Reset)
             begin  -- process csum1_update
               if CPE_Reset = '1' then  -- asynchronous reset (active low)
                 d1 <= (others => "111111");
               elsif CPE_Pclk'event and CPE_Pclk = '1' then  
                  if resetcsumpath = '1' then
                    d1 <= (others => "111111");
                  elsif setcsum_monkey = '1' then
                    for i in 0 to nmax-1 loop
                      if stud(i) = '1' then
                        if valid(i) = '1' then
                          if add1(i) < T then
                            d1(i) <= add1(i);
                          else
                            d1(i) <= (others => '1');
                          end if;
                        end if;
                      end if;
                    end loop;
                  elsif setcsum_asterix = '1' then
                    for i in 0 to nmax-1 loop
                      if honda(i) = '1' then
                        if valid(i) = '1' then
                          if add1(i) < T then
                            d1(i) <= add1(i);
                          else
                            d1(i) <= "111111";
                          end if;
                        end if;
                      end if;
                    end loop;
                  end if;
               end if;
             end process csum1_update;


             path0_update: process (CPE_Pclk, CPE_Reset)
             begin  -- process d0_update
               if CPE_Reset = '1' then  -- asynchronous reset (active low)
                 PathIdentify_0 <= (others => '0');
               elsif CPE_Pclk'event and CPE_Pclk = '1' then  
                  if resetcsumpath = '1' then
                     PathIdentify_0 <= (others => '0');
                  elsif setcsum_sriram = '1' then
                    for i in 0 to nmax-1 loop
                      if stud(i) = '1' then
                        if valid(i) = '1' then
                          if add0(i) < T then
                            PathIdentify_0(i) <= '1';
                          else
                            PathIdentify_0(i) <= '0';
                          end if;
                        end if;
                      end if;
                    end loop;
                  elsif setcsum_tintin = '1' then
                    for i in 0 to nmax-1 loop
                      if honda(i) = '1' then
                        if valid(i) = '1' then
                          if add0(i) < T then
                            PathIdentify_0(i) <= '1';
                          else
                            PathIdentify_0(i) <= '0';
                          end if;
                        end if;
                      end if;
                    end loop;
                  elsif setpath_don = '1' then
                    for buddy in 0 to nmax-1 loop
                      if y(buddy) = '1' then
                        if PathIdentify_0(buddy) = '0' then
                          PathIdentify_0(buddy) <= '1';
                        end if;
                      end if;
                    end loop;
                  end if;
               end if;
             end process path0_update;


              path1_update: process (CPE_Pclk, CPE_Reset)
             begin  -- process d0_update
               if CPE_Reset = '1' then  -- asynchronous reset (active low)
                 PathIdentify_1 <= (others => '0');
               elsif CPE_Pclk'event and CPE_Pclk = '1' then  
                  if resetcsumpath = '1' then
                     PathIdentify_1 <= (others => '0');
                  elsif setcsum_monkey = '1' then
                    for i in 0 to nmax-1 loop
                      if stud(i) = '1' then
                        if valid(i) = '1' then
                          if add1(i) < T then
                            PathIdentify_1(i) <= '1';
                          else
                            PathIdentify_1(i) <= '0';
                          end if;
                        end if;
                      end if;
                    end loop;
                  elsif setcsum_asterix = '1' then
                    for i in 0 to nmax-1 loop
                      if honda(i) = '1' then
                        if valid(i) = '1' then
                          if add1(i) < T then
                            PathIdentify_1(i) <= '1';
                          else
                            PathIdentify_1(i) <= '0';
                          end if;
                        end if;
                      end if;
                    end loop;
                  elsif setpath_don = '1' then
                    for kk in 0 to nmax-1 loop
                      if x(kk) = '1' then
                        PathIdentify_1(kk) <= '0';
                      end if;
                    end loop;
                  end if;
               end if;
             end process path1_update;


             NEXTSTATE0_update: process (CPE_Pclk, CPE_Reset)
             begin  -- process d0_update
               if CPE_Reset = '1' then  -- asynchronous reset (active low)
                 NEXTSTATE0 <= (others => "00000000");
               elsif CPE_Pclk'event and CPE_Pclk = '1' then  
                 if setcsum_sriram = '1' then
                    for i in 0 to nmax-1 loop
                      if stud(i) = '1' then
                        if valid(i) = '1' then
                          if add0(i) < T then
                            NEXTSTATE0(i) <= ns0(i);
                          else
                            NEXTSTATE0(i) <= "00000000";
                          end if;
                        end if;
                      end if;
                    end loop;
                 elsif setcsum_tintin = '1' then
                    for i in 0 to nmax-1 loop
                      if honda(i) = '1' then
                        if valid(i) = '1' then
                          if add0(i) < T then
                            NEXTSTATE0(i) <= ns0(i);
                          else
                            NEXTSTATE0(i) <=  "00000000";
                          end if;
                        end if;
                      end if;
                    end loop;
                 end if;
               end if;
             end process NEXTSTATE0_update;

             NEXTSTATE1_update: process (CPE_Pclk, CPE_Reset)
             begin  -- process d0_update
               if CPE_Reset = '1' then  -- asynchronous reset (active low)
                 NEXTSTATE1 <= (others => "00000000");
               elsif CPE_Pclk'event and CPE_Pclk = '1' then  
                 if setcsum_monkey = '1' then
                    for i in 0 to nmax-1 loop
                      if stud(i) = '1' then
                        if valid(i) = '1' then
                          if add1(i) < T then
                            NEXTSTATE1(i) <= ns1(i);
                          else
                            NEXTSTATE1(i) <=  "00000000";
                          end if;
                        end if;
                      end if;
                    end loop;
                 elsif setcsum_asterix = '1' then
                    for i in 0 to nmax-1 loop
                      if honda(i) = '1' then
                        if valid(i) = '1' then
                          if add1(i) < T then
                            NEXTSTATE1(i) <= ns1(i);
                          else
                            NEXTSTATE1(i) <= "00000000";
                          end if;
                        end if;
                      end if;
                    end loop;
                 end if;
               end if;
             end process NEXTSTATE1_update;
             
             CPE_PeToHostFifoData_Out(35 downto 1) <= (others => '0');
               
             process (present_state,CPE_HosttoPeFifoEmpty_n,ns1,ns0,T,
		      p_metric,stud,level,honda,ll,bb,bun,add1,add0,
		      count,update1,valid,initial,final)

    begin  -- process
           CPE_Fifo_WE_n <= '1';
           CPE_FifoPtrIncr_EN <= '0';
           CPE_PeToHostFifoData_OE <= (others => '0');
           calc_b <= '0';
           inc_T <= '0';
           dec_T <= '0';
           reset_T <= '0';
           inc_level <= '0';
           inc_ll <= '0';
           reset_level <= '0';
           reset_ll <= '0';
           inc_add <= '0';
           inc_honda <= '0';
           reset_honda <= '0';
           reset_stud <= '0';
           inc_stud <= '0';
           reset_count <= '0';
           inc_count <= '0';
           reset_bun <= '0';
           inc_bun <= '0';
           reset_bb <= '0';
           inc_bb <= '0';
           norm_metric <= '0';
           set_out <= '0';
           set_output <= '0'; 
           set_ns <= '0';
           read_ip <= '0';
           set_cumsum_s7 <= '0';
           set_cumsum_s8 <= '0';
           set_cumsum_don <= '0';
           reset_valid <= '0';
           reset_xy <= '0';
           set_x <= '0';
           set_y <= '0';
           set_min1 <= '0';
           set_min2 <= '0';
           set_min3 <= '0';
           set_dm <= '0';
           reset_dm <= '0';
           reset_h <= '0';
           set_h <= '0';
           set_fifo_out <= '0';
           resetcsumpath <= '0';
           setcsum_sriram <= '0';
           setcsum_tintin <= '0';
           setcsum_monkey <= '0';
           setcsum_asterix <= '0';
           setpath_don <= '0';
           reset_row <= '0';
           reset_nextstates <= '0';
           
     case present_state is

       when Read1 => -- Read data from FIFO
  
         if CPE_HosttoPeFifoEmpty_n = '1' then
           CPE_FifoPtrIncr_EN <= '1';
           next_state <= Read2;
         else
           CPE_FifoPtrIncr_EN <= '0';
           next_state <= Read1;
         end if;
         reset_h <= '1';
         -- h <= (others => '0');
         reset_xy <= '1';  

        when Read2 =>
           read_ip <= '1';
           next_state <= s1;
                     
       when s0 => 
           shift(k-2 downto 0) <= (others => '0');
           shift(k-1) <= '1';
           next_state <= Read1;
           reset_level <= '1';
           reset_h <= '1';
           reset_nextstates <= '1';

        when s1 =>
           calc_b <= '1';
           reset_dm <= '1';
           reset_row <= '1';
           next_state <= s2;
           resetcsumpath <= '1';
           set_out <= '1';
           
        when s2 =>
           set_output <= '1';
           set_ns <= '1'; 
           next_state <= s_add;
           inc_add <= '1';
           reset_stud <= '1';
           reset_count <= '1';

       when s_add =>
         next_state <= sriram;
         
       when sriram =>
         setcsum_sriram <= '1';
         for i in 0 to nmax-1 loop
           if stud(i) = '1' then
             if valid(i) = '1' then
               if add0(i) < T then
                 inc_count <= '1';
               else
                 inc_count <= '0';
               end if;
             end if;
           end if;
         end loop;
         next_state <= monkey;
           
       when monkey =>
           setcsum_monkey <= '1';  
           for i in 0 to nmax-1 loop
             if stud(i) = '1' then
               if valid(i) = '1' then
                 if add1(i) < T then
                   inc_count <= '1';
                 else
                   inc_count <= '0';
                 end if;
               end if;                        -- valid
             end if;
           end loop;
           if stud(nmax-1) = '1' then
             next_state <= dummy_s3;
           else
              inc_stud <= '1';
              next_state <= s_add;
           end if;
         
      when dummy_s3 =>
          next_state <= s3;
          
       when s3 =>
         if count > nmax then           -- number of paths compared with nmax
           dec_T <= '1';
           next_state <= s2;            -- if count > nmax reduce T and repeat
                                        -- di < dm + T
         elsif count = 0 then           -- boundary case when count = 0
           inc_T <= '1';
           next_state <= dummy_tintin;
           reset_honda <= '1';
           reset_count <= '1';
         else
           next_state <= s4;
         end if;

       when dummy_tintin =>
           next_state <= tintin;

       when tintin =>
         setcsum_tintin <= '1';
           for i in 0 to nmax-1 loop
             if honda(i) = '1' then
               if valid(i) = '1' then
                 if add0(i) < T then
                   inc_count <= '1';
                 else
                   inc_count <= '0';
                 end if;
               end if;
             end if;
           end loop;
           if honda(nmax-1) = '0' then
             if nmax-2 > count then
               next_state <= asterix;
             else
               next_state <= s4;
             end if;
           end if;
           
       when asterix =>
         setcsum_asterix <= '1';
           for i in 0 to nmax-1 loop
             if honda(i) = '1' then
               if valid(i) = '1' then
                 if add1(i) < T then
                   inc_count <= '1';
                 else
                   inc_count <= '0';
                 end if;
               end if; 
             end if;
           end loop;
           inc_honda <= '1';  
         
         if honda(nmax-1) = '0' then
           if nmax-2 > count then
             next_state <= dummy_tintin;
           else
             next_state <= s4;
           end if;
         else
           next_state <= s4;
         end if;

     -- Determine dm

       when s4 =>
           set_min1 <= '1';
           next_state <= s5;

       when s5 =>
         set_min2 <= '1';
         next_state <= minimum;

       when minimum =>
         set_min3 <= '1';
         next_state <= s6;
                 
       when s6 =>
        set_dm <= '1';
        next_state <= s7;
        reset_valid <= '1';

       when s7 =>
        set_cumsum_s7 <= '1';
        next_state <= dummy_s8;

       when dummy_s8 =>
         next_state <= s8;

       when s8 =>
         next_state <= dummy_dumbo;
	 reset_bun <= '1';
         set_cumsum_s8 <= '1';

         
       when dummy_dumbo =>
           next_state <= dumbo;
         
       when dumbo =>
         if update1 = "00000000" then
           next_state <= stest;
         else
           next_state <= s9;
	   reset_bb <= '1';
         end if;

       when s9 =>                       -- copy memory row if a path exists already
           set_x <= '1';
           for pop in 0 to nmax-1 loop
             if bun(pop) = '1' then
               if update1(pop) = '1' then
                 next_state <= tom;
               else
                 if bun(nmax-1) = '0' then
                   inc_bun <= '1';
                   next_state <= jim;
                 else
                   next_state <= stest;
                   reset_ll <= '1';
                 end if;
               end if;
             end if;
           end loop;  -- pop


       when jim =>
          next_state <= s9;
                      
       when tom => -- Search for a survivor memory row with valid = 0
           set_y <= '1';
           for dolly in 0 to nmax-1 loop
             if bb(dolly) = '1' then
               if valid(dolly) = '0' then
                 next_state <= don;
               else
                 next_state <= bob;
                 inc_bb <= '1';
               end if;
             end if;
           end loop;  -- dolly
          
          --end if;
           
       when bob =>
           --if bb(nmax-1) = '0' then
             next_state <= tom;
           --end if;

       when don =>
           setpath_don <= '1';
           set_cumsum_don <= '1';
           if bun(nmax-1) = '0' then
             next_state <= s9;
             inc_bun <= '1';
             reset_xy <= '1';
           else
             next_state <= stest;
             reset_ll <= '1';
           end if;

       when stest =>
         next_state <= s11;
         set_h <= '1';
               
       when s11 =>

         if initial = '0' then
           set_fifo_out <= '1'; -- output to FIFO
         end if;

	 if (final = '1') then
	     next_state <= dummy;
	 else
	     next_state <= s10;
	 end if;

       when dummy =>
          if level(tl-1) = '0' then
            inc_level <= '1';
            --level := level + 1;
          else
            reset_level <= '1';
            --level := 0;
          end if;
          if initial = '0' then
           CPE_Fifo_WE_n <= '0';
           CPE_FifoPtrIncr_EN <= '1';
           CPE_PeToHostFifoData_OE <= (others => '1');
          end if;

          if (ll < tl - 1) then
             inc_ll <= '1';
	     --ll := ll +1;
	     next_state <= s13;
	 else
             reset_ll <= '1';
             --ll := 0;
	     next_state <= s10;
	 end if;


       when s13 =>
         next_state <= s11;
         
       when s10 =>
         if initial = '0' then
           CPE_Fifo_WE_n <= '0';
           CPE_FifoPtrIncr_EN <= '1';
           CPE_PeToHostFifoData_OE <= (others => '1');
         end if;
         reset_T <= '1';    
         inc_level <= '1';
         reset_xy <= '1';
         reset_h <= '1';
         norm_metric <=  '1'; -- Normalize path metrics at each trellis stage
         if final = '1' then
           next_state <= dummy_s12;
         else
           next_state <= Read1;
         end if;

       when dummy_s12 => -- dummy states, goes in case of exceptions
         next_state <= s12;  
             
       when s12 =>
         next_state <= s12;

       when others =>
         next_state <= s0;
     end case;    

 end process;

end behavior;
